@@ -1,6 +1,6 @@
 
 /*
- * $Id: delay_pools.c,v 1.7.2.3 1999/04/08 07:38:32 wessels Exp $
+ * $Id: delay_pools.c,v 1.17 2000/05/16 07:06:04 wessels Exp $
  *
  * DEBUG: section 77    Delay Pools
  * AUTHOR: David Luyer <luyer@ucs.uwa.edu.au>
@@ -12,10 +12,10 @@
  *  Internet community.  Development is led by Duane Wessels of the
  *  National Laboratory for Applied Network Research and funded by the
  *  National Science Foundation.  Squid is Copyrighted (C) 1998 by
- *  Duane Wessels and the University of California San Diego.  Please
- *  see the COPYRIGHT file for full details.  Squid incorporates
- *  software developed and/or copyrighted by other sources.  Please see
- *  the CREDITS file for full details.
+ *  the Regents of the University of California.  Please see the
+ *  COPYRIGHT file for full details.  Squid incorporates software
+ *  developed and/or copyrighted by other sources.  Please see the
+ *  CREDITS file for full details.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -39,33 +39,41 @@
 #include "squid.h"
 
 struct _class1DelayPool {
+    int class;
     int aggregate;
 };
 
+#define IND_MAP_SZ 256
+
 struct _class2DelayPool {
+    int class;
     int aggregate;
     /* OK: -1 is terminator.  individual[255] is always host 255. */
     /* 255 entries + 1 terminator byte */
-    unsigned char individual_map[256];
+    unsigned char individual_map[IND_MAP_SZ];
     unsigned char individual_255_used;
     /* 256 entries */
-    int individual[256];
+    int individual[IND_MAP_SZ];
 };
 
+#define NET_MAP_SZ 256
+#define C3_IND_SZ (NET_MAP_SZ*IND_MAP_SZ)
+
 struct _class3DelayPool {
+    int class;
     int aggregate;
     /* OK: -1 is terminator.  network[255] is always host 255. */
     /* 255 entries + 1 terminator byte */
-    unsigned char network_map[256];
+    unsigned char network_map[NET_MAP_SZ];
     unsigned char network_255_used;
     /* 256 entries */
     int network[256];
     /* 256 sets of (255 entries + 1 terminator byte) */
-    unsigned char individual_map[256][256];
+    unsigned char individual_map[NET_MAP_SZ][IND_MAP_SZ];
     /* Pack this into one bit per net */
     unsigned char individual_255_used[32];
     /* largest entry = (255<<8)+255 = 65535 */
-    int individual[65536];
+    int individual[C3_IND_SZ];
 };
 
 typedef struct _class1DelayPool class1DelayPool;
@@ -84,6 +92,7 @@ static delayPool *delay_data = NULL;
 static fd_set delay_no_delay;
 static time_t delay_pools_last_update = 0;
 static hash_table *delay_id_ptr_hash = NULL;
+static long memory_used = 0;
 
 static OBJH delayPoolStats;
 
@@ -113,8 +122,12 @@ delayIdPtrHash(const void *key, unsigned int n)
 static int
 delayIdPtrHashCmp(const void *a, const void *b)
 {
-    /* Sort by POINTER VALUE. */
-    return b - a;
+    /*
+     * Compare POINTER VALUE.
+     * Note, we can't subtract void pointers, but we don't need
+     * to anyway.  All we need is a test for equality.
+     */
+    return a != b;
 }
 
 void
@@ -128,37 +141,46 @@ delayPoolsInit(void)
 void
 delayInitDelayData(unsigned short pools)
 {
-    if (pools) {
-	delay_data = xcalloc(pools, sizeof(delayPool));
-	eventAdd("delayPoolsUpdate", delayPoolsUpdate, NULL, 1.0, 1);
-	delay_id_ptr_hash = hash_create(delayIdPtrHashCmp, 256, delayIdPtrHash);
-    }
+    if (!pools)
+	return;
+    delay_data = xcalloc(pools, sizeof(*delay_data));
+    memory_used += sizeof(*delay_data);
+    eventAdd("delayPoolsUpdate", delayPoolsUpdate, NULL, 1.0, 1);
+    delay_id_ptr_hash = hash_create(delayIdPtrHashCmp, 256, delayIdPtrHash);
 }
 
 static void
 delayIdZero(void *hlink)
 {
     hash_link *h = hlink;
-    *(delay_id *) (h->key) = 0;
+    delay_id *id = (delay_id *) h->key;
+    *id = 0;
     xfree(h);
+    memory_used -= sizeof(*h);
 }
 
 void
-delayFreeDelayData()
+delayFreeDelayData(void)
 {
     safe_free(delay_data);
-    if (delay_id_ptr_hash) {
-	hashFreeItems(delay_id_ptr_hash, delayIdZero);
-	hashFreeMemory(delay_id_ptr_hash);
-	delay_id_ptr_hash = NULL;
-    }
+    memory_used -= sizeof(*delay_data);
+    if (!delay_id_ptr_hash)
+	return;
+    hashFreeItems(delay_id_ptr_hash, delayIdZero);
+    hashFreeMemory(delay_id_ptr_hash);
+    delay_id_ptr_hash = NULL;
 }
 
 void
 delayRegisterDelayIdPtr(delay_id * loc)
 {
-    hash_link *lnk = xmalloc(sizeof(hash_link));
-
+    hash_link *lnk;
+    if (!delay_id_ptr_hash)
+	return;
+    if (*loc == 0)
+	return;
+    lnk = xmalloc(sizeof(hash_link));
+    memory_used += sizeof(hash_link);
     lnk->key = (char *) loc;
     hash_join(delay_id_ptr_hash, lnk);
 }
@@ -166,11 +188,21 @@ delayRegisterDelayIdPtr(delay_id * loc)
 void
 delayUnregisterDelayIdPtr(delay_id * loc)
 {
-    hash_link *lnk = hash_lookup(delay_id_ptr_hash, loc);
-
+    hash_link *lnk;
+    if (!delay_id_ptr_hash)
+	return;
+    /*
+     * If we went through a reconfigure, then all the delay_id's
+     * got set to zero, and they were removed from our hash
+     * table.
+     */
+    if (*loc == 0)
+	return;
+    lnk = hash_lookup(delay_id_ptr_hash, loc);
     assert(lnk);
     hash_remove_link(delay_id_ptr_hash, lnk);
     xxfree(lnk);
+    memory_used -= sizeof(*lnk);
 }
 
 void
@@ -179,12 +211,18 @@ delayCreateDelayPool(unsigned short pool, u_char class)
     switch (class) {
     case 1:
 	delay_data[pool].class1 = xmalloc(sizeof(class1DelayPool));
+	delay_data[pool].class1->class = 1;
+	memory_used += sizeof(class1DelayPool);
 	break;
     case 2:
 	delay_data[pool].class2 = xmalloc(sizeof(class2DelayPool));
+	delay_data[pool].class1->class = 2;
+	memory_used += sizeof(class2DelayPool);
 	break;
     case 3:
 	delay_data[pool].class3 = xmalloc(sizeof(class3DelayPool));
+	delay_data[pool].class1->class = 3;
+	memory_used += sizeof(class3DelayPool);
 	break;
     default:
 	assert(0);
@@ -199,18 +237,18 @@ delayInitDelayPool(unsigned short pool, u_char class, delaySpecSet * rates)
      */
     switch (class) {
     case 1:
-	delay_data[pool].class1->aggregate = (rates->aggregate.max_bytes *
-	    Config.Delay.initial) / 100;
+	delay_data[pool].class1->aggregate = (int) (((double) rates->aggregate.max_bytes *
+		Config.Delay.initial) / 100);
 	break;
     case 2:
-	delay_data[pool].class2->aggregate = (rates->aggregate.max_bytes *
-	    Config.Delay.initial) / 100;
+	delay_data[pool].class2->aggregate = (int) (((double) rates->aggregate.max_bytes *
+		Config.Delay.initial) / 100);
 	delay_data[pool].class2->individual_map[0] = 255;
 	delay_data[pool].class2->individual_255_used = 0;
 	break;
     case 3:
-	delay_data[pool].class3->aggregate = (rates->aggregate.max_bytes *
-	    Config.Delay.initial) / 100;
+	delay_data[pool].class3->aggregate = (int) (((double) rates->aggregate.max_bytes *
+		Config.Delay.initial) / 100);
 	delay_data[pool].class3->network_map[0] = 255;
 	delay_data[pool].class3->network_255_used = 0;
 	memset(&delay_data[pool].class3->individual_255_used, '\0',
@@ -225,6 +263,20 @@ void
 delayFreeDelayPool(unsigned short pool)
 {
     /* this is a union - and all free() cares about is the pointer location */
+    switch (delay_data[pool].class1->class) {
+    case 1:
+	memory_used -= sizeof(class1DelayPool);
+	break;
+    case 2:
+	memory_used -= sizeof(class2DelayPool);
+	break;
+    case 3:
+	memory_used -= sizeof(class3DelayPool);
+	break;
+    default:
+	debug(77, 1) ("delayFreeDelayPool: bad class %d\n",
+	    delay_data[pool].class1->class);
+    }
     safe_free(delay_data[pool].class1);
 }
 
@@ -265,7 +317,12 @@ delayClient(request_t * r)
     memset(&ch, '\0', sizeof(ch));
     ch.src_addr = r->client_addr;
     ch.my_addr = r->my_addr;
+    ch.my_port = r->my_port;
     ch.request = r;
+    if (r->client_addr.s_addr == INADDR_BROADCAST) {
+	debug(77, 2) ("delayClient: WARNING: Called with 'allones' address, ignoring\n");
+	return delayId(0, 0);
+    }
     for (pool = 0; pool < Config.Delay.pools; pool++) {
 	if (aclCheckFast(Config.Delay.access[pool], &ch))
 	    break;
@@ -282,21 +339,22 @@ delayClient(request_t * r)
 	if (host == 255) {
 	    if (!delay_data[pool].class2->individual_255_used) {
 		delay_data[pool].class2->individual_255_used = 1;
-		delay_data[pool].class2->individual[255] =
-		    (Config.Delay.rates[pool]->individual.max_bytes *
-		    Config.Delay.initial) / 100;
+		delay_data[pool].class2->individual[IND_MAP_SZ - 1] =
+		    (int) (((double) Config.Delay.rates[pool]->individual.max_bytes *
+			Config.Delay.initial) / 100);
 	    }
 	    return delayId(pool + 1, 255);
 	}
-	for (i = 0;; i++) {
+	for (i = 0; i < IND_MAP_SZ; i++) {
 	    if (delay_data[pool].class2->individual_map[i] == host)
 		break;
 	    if (delay_data[pool].class2->individual_map[i] == 255) {
 		delay_data[pool].class2->individual_map[i] = host;
+		assert(i < (IND_MAP_SZ - 1));
 		delay_data[pool].class2->individual_map[i + 1] = 255;
 		delay_data[pool].class2->individual[i] =
-		    (Config.Delay.rates[pool]->individual.max_bytes *
-		    Config.Delay.initial) / 100;
+		    (int) (((double) Config.Delay.rates[pool]->individual.max_bytes *
+			Config.Delay.initial) / 100);
 		break;
 	    }
 	}
@@ -311,20 +369,21 @@ delayClient(request_t * r)
 	if (!delay_data[pool].class3->network_255_used) {
 	    delay_data[pool].class3->network_255_used = 1;
 	    delay_data[pool].class3->network[255] =
-		(Config.Delay.rates[pool]->network.max_bytes *
-		Config.Delay.initial) / 100;
+		(int) (((double) Config.Delay.rates[pool]->network.max_bytes *
+		    Config.Delay.initial) / 100);
 	}
     } else {
-	for (i = 0;; i++) {
+	for (i = 0; i < NET_MAP_SZ; i++) {
 	    if (delay_data[pool].class3->network_map[i] == net)
 		break;
 	    if (delay_data[pool].class3->network_map[i] == 255) {
 		delay_data[pool].class3->network_map[i] = net;
 		delay_data[pool].class3->individual_map[i][0] = 255;
+		assert(i < (NET_MAP_SZ - 1));
 		delay_data[pool].class3->network_map[i + 1] = 255;
 		delay_data[pool].class3->network[i] =
-		    (Config.Delay.rates[pool]->network.max_bytes *
-		    Config.Delay.initial) / 100;
+		    (int) (((double) Config.Delay.rates[pool]->network.max_bytes *
+			Config.Delay.initial) / 100);
 		break;
 	    }
 	}
@@ -334,23 +393,28 @@ delayClient(request_t * r)
 	position |= 255;
 	if (!(delay_data[pool].class3->individual_255_used[i / 8] & (1 << (i % 8)))) {
 	    delay_data[pool].class3->individual_255_used[i / 8] |= (1 << (i % 8));
+	    assert(position < C3_IND_SZ);
 	    delay_data[pool].class3->individual[position] =
-		(Config.Delay.rates[pool]->individual.max_bytes *
-		Config.Delay.initial) / 100;
+		(int) (((double) Config.Delay.rates[pool]->individual.max_bytes *
+		    Config.Delay.initial) / 100);
 	}
 	return delayId(pool + 1, position);
     }
-    for (j = 0;; j++) {
+    assert(i < NET_MAP_SZ);
+    for (j = 0; j < IND_MAP_SZ; j++) {
 	if (delay_data[pool].class3->individual_map[i][j] == host) {
 	    position |= j;
 	    break;
 	}
 	if (delay_data[pool].class3->individual_map[i][j] == 255) {
 	    delay_data[pool].class3->individual_map[i][j] = host;
+	    assert(j < (IND_MAP_SZ - 1));
 	    delay_data[pool].class3->individual_map[i][j + 1] = 255;
-	    delay_data[pool].class3->individual[position |= j] =
-		(Config.Delay.rates[pool]->individual.max_bytes *
-		Config.Delay.initial) / 100;
+	    position |= j;
+	    assert(position < C3_IND_SZ);
+	    delay_data[pool].class3->individual[position] =
+		(int) (((double) Config.Delay.rates[pool]->individual.max_bytes *
+		    Config.Delay.initial) / 100);
 	    break;
 	}
     }
@@ -373,7 +437,7 @@ static void
 delayUpdateClass2(class2DelayPool * class2, delaySpecSet * rates, int incr)
 {
     int restore_bytes;
-    unsigned char i;
+    unsigned int i;
     /* delaySetSpec may be pointer to partial structure so MUST pass by
      * reference.
      */
@@ -389,12 +453,13 @@ delayUpdateClass2(class2DelayPool * class2, delaySpecSet * rates, int incr)
     else
 	i = 0;
     for (;;) {
+	assert(i < IND_MAP_SZ);
 	if (i != 255 && class2->individual_map[i] == 255)
 	    return;
 	if (class2->individual[i] != rates->individual.max_bytes &&
 	    (class2->individual[i] += restore_bytes) > rates->individual.max_bytes)
 	    class2->individual[i] = rates->individual.max_bytes;
-	if (++i == 255)
+	if (++i == (IND_MAP_SZ - 1))
 	    return;
     }
 }
@@ -404,7 +469,8 @@ delayUpdateClass3(class3DelayPool * class3, delaySpecSet * rates, int incr)
 {
     int individual_restore_bytes, network_restore_bytes;
     int mpos;
-    unsigned char i, j;
+    unsigned int i, j;
+    char individual_255_used;
     /* delaySetSpec may be pointer to partial structure so MUST pass by
      * reference.
      */
@@ -420,31 +486,23 @@ delayUpdateClass3(class3DelayPool * class3, delaySpecSet * rates, int incr)
 	return;
     individual_restore_bytes *= incr;
     network_restore_bytes *= incr;
-    if (class3->network_255_used)
-	i = 255;
-    else
-	i = 0;
-    for (;;) {
+    for (i = 0; i < ((class3->network_255_used) ? NET_MAP_SZ : NET_MAP_SZ - 1); ++i) {
+	assert(i < NET_MAP_SZ);
 	if (i != 255 && class3->network_map[i] == 255)
 	    return;
 	if (individual_restore_bytes != -incr) {
 	    mpos = i << 8;
-	    if (class3->individual_255_used[i / 8] & (1 << (i % 8)))
-		j = 255;
-	    else
-		j = 0;
-	    for (;;) {
+	    individual_255_used = class3->individual_255_used[i / 8] & (1 << (i % 8));
+	    for (j = 0; j < ((individual_255_used) ? IND_MAP_SZ : IND_MAP_SZ - 1); ++j, ++mpos) {
+		assert(i < NET_MAP_SZ);
+		assert(j < IND_MAP_SZ);
 		if (j != 255 && class3->individual_map[i][j] == 255)
 		    break;
+		assert(mpos < C3_IND_SZ);
 		if (class3->individual[mpos] != rates->individual.max_bytes &&
 		    (class3->individual[mpos] += individual_restore_bytes) >
 		    rates->individual.max_bytes)
 		    class3->individual[mpos] = rates->individual.max_bytes;
-		mpos++;
-		if (j == 255)
-		    mpos -= 256;
-		if (++j == 255)
-		    break;
 	    }
 	}
 	if (network_restore_bytes != -incr &&
@@ -452,7 +510,7 @@ delayUpdateClass3(class3DelayPool * class3, delaySpecSet * rates, int incr)
 	    (class3->network[i] += network_restore_bytes) >
 	    rates->network.max_bytes)
 	    class3->network[i] = rates->network.max_bytes;
-	if (++i == 255)
+	if (++i == (NET_MAP_SZ - 1))
 	    return;
     }
 }
@@ -573,7 +631,9 @@ delayMostBytesWanted(const MemObject * mem, int max)
     int i = 0;
     int found = 0;
     store_client *sc;
-    for (sc = mem->clients; sc; sc = sc->next) {
+    dlink_node *node;
+    for (node = mem->clients.head; node; node = node->next) {
+	sc = (store_client *) node->data;
 	if (sc->callback_data == NULL)	/* open slot */
 	    continue;
 	if (sc->type != STORE_MEM_CLIENT)
@@ -590,8 +650,10 @@ delayMostBytesAllowed(const MemObject * mem)
     int j;
     int jmax = -1;
     store_client *sc;
+    dlink_node *node;
     delay_id d = 0;
-    for (sc = mem->clients; sc; sc = sc->next) {
+    for (node = mem->clients.head; node; node = node->next) {
+	sc = (store_client *) node->data;
 	if (sc->callback_data == NULL)	/* open slot */
 	    continue;
 	if (sc->type != STORE_MEM_CLIENT)
@@ -606,9 +668,8 @@ delayMostBytesAllowed(const MemObject * mem)
 }
 
 void
-delaySetStoreClient(StoreEntry * e, void *data, delay_id delay_id)
+delaySetStoreClient(store_client * sc, delay_id delay_id)
 {
-    store_client *sc = storeClientListSearch(e->mem_obj, data);
     assert(sc != NULL);
     sc->delay_id = delay_id;
     delayRegisterDelayIdPtr(&sc->delay_id);
@@ -644,7 +705,8 @@ delayPoolStats2(StoreEntry * sentry, unsigned short pool)
     /* must be a reference only - partially malloc()d struct */
     delaySpecSet *rate = Config.Delay.rates[pool];
     class2DelayPool *class2 = delay_data[pool].class2;
-    unsigned char shown = 0, i;
+    unsigned char shown = 0;
+    unsigned int i;
 
     storeAppendPrintf(sentry, "Pool: %d\n\tClass: 2\n\n", pool + 1);
     delayPoolStatsAg(sentry, rate, class2->aggregate);
@@ -656,7 +718,7 @@ delayPoolStats2(StoreEntry * sentry, unsigned short pool)
     storeAppendPrintf(sentry, "\t\tMax: %d\n", rate->individual.max_bytes);
     storeAppendPrintf(sentry, "\t\tRate: %d\n", rate->individual.restore_bps);
     storeAppendPrintf(sentry, "\t\tCurrent: ");
-    for (i = 0;; i++) {
+    for (i = 0; i < IND_MAP_SZ; i++) {
 	if (class2->individual_map[i] == 255)
 	    break;
 	storeAppendPrintf(sentry, "%d:%d ", class2->individual_map[i],
@@ -678,7 +740,9 @@ delayPoolStats3(StoreEntry * sentry, unsigned short pool)
     /* fully malloc()d struct in this case only */
     delaySpecSet *rate = Config.Delay.rates[pool];
     class3DelayPool *class3 = delay_data[pool].class3;
-    unsigned char shown = 0, i, j;
+    unsigned char shown = 0;
+    unsigned int i;
+    unsigned int j;
 
     storeAppendPrintf(sentry, "Pool: %d\n\tClass: 3\n\n", pool + 1);
     delayPoolStatsAg(sentry, rate, class3->aggregate);
@@ -689,7 +753,7 @@ delayPoolStats3(StoreEntry * sentry, unsigned short pool)
 	storeAppendPrintf(sentry, "\t\tMax: %d\n", rate->network.max_bytes);
 	storeAppendPrintf(sentry, "\t\tRate: %d\n", rate->network.restore_bps);
 	storeAppendPrintf(sentry, "\t\tCurrent: ");
-	for (i = 0;; i++) {
+	for (i = 0; i < NET_MAP_SZ; i++) {
 	    if (class3->network_map[i] == 255)
 		break;
 	    storeAppendPrintf(sentry, "%d:%d ", class3->network_map[i],
@@ -712,12 +776,12 @@ delayPoolStats3(StoreEntry * sentry, unsigned short pool)
     storeAppendPrintf(sentry, "\tIndividual:\n");
     storeAppendPrintf(sentry, "\t\tMax: %d\n", rate->individual.max_bytes);
     storeAppendPrintf(sentry, "\t\tRate: %d\n", rate->individual.restore_bps);
-    for (i = 0;; i++) {
+    for (i = 0; i < NET_MAP_SZ; i++) {
 	if (class3->network_map[i] == 255)
 	    break;
 	storeAppendPrintf(sentry, "\t\tCurrent [Network %d]: ", class3->network_map[i]);
 	shown = 1;
-	for (j = 0;; j++) {
+	for (j = 0; j < IND_MAP_SZ; j++) {
 	    if (class3->individual_map[i][j] == 255)
 		break;
 	    storeAppendPrintf(sentry, "%d:%d ", class3->individual_map[i][j],
@@ -731,7 +795,7 @@ delayPoolStats3(StoreEntry * sentry, unsigned short pool)
     if (class3->network_255_used) {
 	storeAppendPrintf(sentry, "\t\tCurrent [Network 255]: ");
 	shown = 1;
-	for (j = 0;; j++) {
+	for (j = 0; j < IND_MAP_SZ; j++) {
 	    if (class3->individual_map[255][j] == 255)
 		break;
 	    storeAppendPrintf(sentry, "%d:%d ", class3->individual_map[255][j],
@@ -772,6 +836,7 @@ delayPoolStats(StoreEntry * sentry)
 	    assert(0);
 	}
     }
+    storeAppendPrintf(sentry, "Memory Used: %d bytes\n", (int) memory_used);
 }
 
 #endif
