@@ -1,6 +1,6 @@
 
 /*
- * $Id: store_client.c,v 1.59.2.4 1999/03/23 05:29:19 wessels Exp $
+ * $Id: store_client.c,v 1.95 2000/06/27 22:06:04 hno Exp $
  *
  * DEBUG: section 20    Storage Manager Client-Side Interface
  * AUTHOR: Duane Wessels
@@ -12,10 +12,10 @@
  *  Internet community.  Development is led by Duane Wessels of the
  *  National Laboratory for Applied Network Research and funded by the
  *  National Science Foundation.  Squid is Copyrighted (C) 1998 by
- *  Duane Wessels and the University of California San Diego.  Please
- *  see the COPYRIGHT file for full details.  Squid incorporates
- *  software developed and/or copyrighted by other sources.  Please see
- *  the CREDITS file for full details.
+ *  the Regents of the University of California.  Please see the
+ *  COPYRIGHT file for full details.  Squid incorporates software
+ *  developed and/or copyrighted by other sources.  Please see the
+ *  CREDITS file for full details.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -40,10 +40,10 @@
  *       'Body' refers to the swapfile body, which is the full
  *        HTTP reply (including HTTP headers and body).
  */
-static DRCB storeClientReadBody;
-static DRCB storeClientReadHeader;
-static SIH storeClientFileOpened;
+static STRCB storeClientReadBody;
+static STRCB storeClientReadHeader;
 static void storeClientCopy2(StoreEntry * e, store_client * sc);
+static void storeClientCopy3(StoreEntry * e, store_client * sc);
 static void storeClientFileRead(store_client * sc);
 static EVH storeClientCopyEvent;
 static store_client_t storeClientType(StoreEntry *);
@@ -56,24 +56,30 @@ int
 storeClientWaiting(const StoreEntry * e)
 {
     MemObject *mem = e->mem_obj;
+    dlink_node *node;
     store_client *sc;
-    for (sc = mem->clients; sc; sc = sc->next) {
+    for (node = mem->clients.head; node; node = node->next) {
+	sc = node->data;
 	if (sc->callback_data != NULL)
 	    return 1;
     }
     return 0;
 }
 
+#if STORE_CLIENT_LIST_DEBUG
 store_client *
 storeClientListSearch(const MemObject * mem, void *data)
 {
-    store_client *sc;
-    for (sc = mem->clients; sc; sc = sc->next) {
+    dlink_node *node;
+    store_client *sc = NULL;
+    for (node = mem->clients.head; node; node = node->next) {
+	sc = node->data;
 	if (sc->callback_data == data)
-	    break;
+	    return sc;
     }
-    return sc;
+    return NULL;
 }
+#endif
 
 static store_client_t
 storeClientType(StoreEntry * e)
@@ -98,6 +104,14 @@ storeClientType(StoreEntry * e)
      */
     else if (mem->nclients == 1)
 	return STORE_MEM_CLIENT;
+    /*
+     * If there is no disk file to open yet, we must make this a
+     * mem client.  If we can't open the swapin file before writing
+     * to the client, there is no guarantee that we will be able
+     * to open it later when we really need it.
+     */
+    else if (e->swap_status == SWAPOUT_NONE)
+	return STORE_MEM_CLIENT;
     /*
      * otherwise, make subsequent clients read from disk so they
      * can not delay the first, and vice-versa.
@@ -107,31 +121,48 @@ storeClientType(StoreEntry * e)
 }
 
 /* add client with fd to client list */
-void
+store_client *
 storeClientListAdd(StoreEntry * e, void *data)
 {
     MemObject *mem = e->mem_obj;
-    store_client **T;
     store_client *sc;
     assert(mem);
+#if STORE_CLIENT_LIST_DEBUG
     if (storeClientListSearch(mem, data) != NULL)
-	return;
+	assert(1 == 0);		/* XXX die! */
+#endif
+    e->refcount++;
     mem->nclients++;
     sc = memAllocate(MEM_STORE_CLIENT);
     cbdataAdd(sc, memFree, MEM_STORE_CLIENT);	/* sc is callback_data for file_read */
+    cbdataLock(data);		/* locked while we point to it */
     sc->callback_data = data;
     sc->seen_offset = 0;
     sc->copy_offset = 0;
-    sc->swapin_fd = -1;
     sc->flags.disk_io_pending = 0;
     sc->entry = e;
     sc->type = storeClientType(e);
     if (sc->type == STORE_DISK_CLIENT)
 	/* assert we'll be able to get the data we want */
 	/* maybe we should open swapin_fd here */
-	assert(e->swap_file_number > -1 || storeSwapOutAble(e));
-    for (T = &mem->clients; *T; T = &(*T)->next);
-    *T = sc;
+	assert(e->swap_filen > -1 || storeSwapOutAble(e));
+    dlinkAdd(sc, &sc->node, &mem->clients);
+#if DELAY_POOLS
+    sc->delay_id = 0;
+#endif
+    return sc;
+}
+
+static void
+storeClientCallback(store_client * sc, ssize_t sz)
+{
+    STCB *callback = sc->callback;
+    char *buf = sc->copy_buf;
+    assert(sc->callback);
+    sc->callback = NULL;
+    sc->copy_buf = NULL;
+    if (cbdataValid(sc->callback_data))
+	callback(sc->callback_data, buf, sz);
 }
 
 static void
@@ -147,7 +178,8 @@ storeClientCopyEvent(void *data)
 
 /* copy bytes requested by the client */
 void
-storeClientCopy(StoreEntry * e,
+storeClientCopy(store_client * sc,
+    StoreEntry * e,
     off_t seen_offset,
     off_t copy_offset,
     size_t size,
@@ -155,7 +187,6 @@ storeClientCopy(StoreEntry * e,
     STCB * callback,
     void *data)
 {
-    store_client *sc;
     assert(!EBIT_TEST(e->flags, ENTRY_ABORTED));
     debug(20, 3) ("storeClientCopy: %s, seen %d, want %d, size %d, cb %p, cbdata %p\n",
 	storeKeyText(e->key),
@@ -164,9 +195,12 @@ storeClientCopy(StoreEntry * e,
 	(int) size,
 	callback,
 	data);
-    sc = storeClientListSearch(e->mem_obj, data);
     assert(sc != NULL);
+#if STORE_CLIENT_LIST_DEBUG
+    assert(sc == storeClientListSearch(e->mem_obj, data));
+#endif
     assert(sc->callback == NULL);
+    assert(sc->entry == e);
     sc->copy_offset = copy_offset;
     sc->seen_offset = seen_offset;
     sc->callback = callback;
@@ -201,9 +235,6 @@ storeClientNoMoreToSend(StoreEntry * e, store_client * sc)
 static void
 storeClientCopy2(StoreEntry * e, store_client * sc)
 {
-    STCB *callback = sc->callback;
-    MemObject *mem = e->mem_obj;
-    size_t sz;
     if (sc->flags.copy_event_pending)
 	return;
     if (EBIT_TEST(e->flags, ENTRY_FWD_HDR_WAIT)) {
@@ -219,7 +250,7 @@ storeClientCopy2(StoreEntry * e, store_client * sc)
     cbdataLock(sc);		/* ick, prevent sc from getting freed */
     sc->flags.store_copying = 1;
     debug(20, 3) ("storeClientCopy2: %s\n", storeKeyText(e->key));
-    assert(callback != NULL);
+    assert(sc->callback != NULL);
     /*
      * We used to check for ENTRY_ABORTED here.  But there were some
      * problems.  For example, we might have a slow client (or two) and
@@ -227,78 +258,73 @@ storeClientCopy2(StoreEntry * e, store_client * sc)
      * if the server-side aborts, we want to give the client(s)
      * everything we got before the abort condition occurred.
      */
-    if (storeClientNoMoreToSend(e, sc)) {
-	/* There is no more to send! */
-#if USE_ASYNC_IO
-	if (sc->flags.disk_io_pending) {
-	    if (sc->swapin_fd >= 0)
-		aioCancel(sc->swapin_fd, NULL);
-	    else
-		aioCancel(-1, sc);
-	}
-#endif
-	sc->flags.disk_io_pending = 0;
-	sc->callback = NULL;
-	callback(sc->callback_data, sc->copy_buf, 0);
-    } else if (e->store_status == STORE_PENDING && sc->seen_offset >= mem->inmem_hi) {
-	/* client has already seen this, wait for more */
-	debug(20, 3) ("storeClientCopy2: Waiting for more\n");
-    } else if (sc->copy_offset >= mem->inmem_lo && sc->copy_offset < mem->inmem_hi) {
-	/* What the client wants is in memory */
-	debug(20, 3) ("storeClientCopy2: Copying from memory\n");
-	sz = stmemCopy(&mem->data_hdr, sc->copy_offset, sc->copy_buf, sc->copy_size);
-#if USE_ASYNC_IO
-	if (sc->flags.disk_io_pending) {
-	    if (sc->swapin_fd >= 0)
-		aioCancel(sc->swapin_fd, NULL);
-	    else
-		aioCancel(-1, sc);
-	}
-#endif
-	sc->flags.disk_io_pending = 0;
-	sc->callback = NULL;
-	callback(sc->callback_data, sc->copy_buf, sz);
-    } else if (sc->swapin_fd < 0) {
-	debug(20, 3) ("storeClientCopy2: Need to open swap in file\n");
-	assert(sc->type == STORE_DISK_CLIENT);
-	/* gotta open the swapin file */
-	if (storeTooManyDiskFilesOpen()) {
-	    /* yuck -- this causes a TCP_SWAPFAIL_MISS on the client side */
-	    sc->callback = NULL;
-	    callback(sc->callback_data, sc->copy_buf, -1);
-	} else if (!sc->flags.disk_io_pending) {
-	    sc->flags.disk_io_pending = 1;
-	    storeSwapInStart(e, storeClientFileOpened, sc);
-	} else {
-	    debug(20, 2) ("storeClientCopy2: Averted multiple fd operation\n");
-	}
-    } else {
-	debug(20, 3) ("storeClientCopy: reading from disk FD %d\n",
-	    sc->swapin_fd);
-	assert(sc->type == STORE_DISK_CLIENT);
-	if (!sc->flags.disk_io_pending) {
-	    storeClientFileRead(sc);
-	} else {
-	    debug(20, 2) ("storeClientCopy2: Averted multiple fd operation\n");
-	}
-    }
+    storeClientCopy3(e, sc);
     sc->flags.store_copying = 0;
     cbdataUnlock(sc);		/* ick, allow sc to be freed */
 }
 
 static void
-storeClientFileOpened(int fd, void *data)
+storeClientCopy3(StoreEntry * e, store_client * sc)
 {
-    store_client *sc = data;
-    STCB *callback = sc->callback;
-    if (fd < 0) {
-	debug(20, 3) ("storeClientFileOpened: failed\n");
-	sc->flags.disk_io_pending = 0;
-	sc->callback = NULL;
-	callback(sc->callback_data, sc->copy_buf, -1);
+    MemObject *mem = e->mem_obj;
+    size_t sz;
+
+    if (storeClientNoMoreToSend(e, sc)) {
+	/* There is no more to send! */
+	storeClientCallback(sc, 0);
 	return;
     }
-    sc->swapin_fd = fd;
+    if (e->store_status == STORE_PENDING && sc->seen_offset >= mem->inmem_hi) {
+	/* client has already seen this, wait for more */
+	debug(20, 3) ("storeClientCopy3: Waiting for more\n");
+	return;
+    }
+    /*
+     * Slight weirdness here.  We open a swapin file for any
+     * STORE_DISK_CLIENT, even if we can copy the requested chunk
+     * from memory in the next block.  We must try to open the
+     * swapin file before sending any data to the client side.  If
+     * we postpone the open, and then can not open the file later
+     * on, the client loses big time.  Its transfer just gets cut
+     * off.  Better to open it early (while the client side handler
+     * is clientCacheHit) so that we can fall back to a cache miss
+     * if needed.
+     */
+    if (STORE_DISK_CLIENT == sc->type && NULL == sc->swapin_sio) {
+	debug(20, 3) ("storeClientCopy3: Need to open swap in file\n");
+	/* gotta open the swapin file */
+	if (storeTooManyDiskFilesOpen()) {
+	    /* yuck -- this causes a TCP_SWAPFAIL_MISS on the client side */
+	    storeClientCallback(sc, -1);
+	    return;
+	} else if (!sc->flags.disk_io_pending) {
+	    /* Don't set store_io_pending here */
+	    storeSwapInStart(sc);
+	    if (NULL == sc->swapin_sio) {
+		storeClientCallback(sc, -1);
+		return;
+	    }
+	    /*
+	     * If the open succeeds we either copy from memory, or
+	     * schedule a disk read in the next block.
+	     */
+	} else {
+	    debug(20, 1) ("WARNING: Averted multiple fd operation (1)\n");
+	    return;
+	}
+    }
+    if (sc->copy_offset >= mem->inmem_lo && sc->copy_offset < mem->inmem_hi) {
+	/* What the client wants is in memory */
+	debug(20, 3) ("storeClientCopy3: Copying from memory\n");
+	sz = stmemCopy(&mem->data_hdr,
+	    sc->copy_offset, sc->copy_buf, sc->copy_size);
+	storeClientCallback(sc, sz);
+	return;
+    }
+    /* What the client wants is not in memory. Schedule a disk read */
+    assert(STORE_DISK_CLIENT == sc->type);
+    assert(!sc->flags.disk_io_pending);
+    debug(20, 3) ("storeClientCopy3: reading from STORE\n");
     storeClientFileRead(sc);
 }
 
@@ -307,11 +333,10 @@ storeClientFileRead(store_client * sc)
 {
     MemObject *mem = sc->entry->mem_obj;
     assert(sc->callback != NULL);
-#ifdef OPTIMISTIC_IO
+    assert(!sc->flags.disk_io_pending);
     sc->flags.disk_io_pending = 1;
-#endif
     if (mem->swap_hdr_sz == 0) {
-	file_read(sc->swapin_fd,
+	storeRead(sc->swapin_sio,
 	    sc->copy_buf,
 	    sc->copy_size,
 	    0,
@@ -319,68 +344,102 @@ storeClientFileRead(store_client * sc)
 	    sc);
     } else {
 	if (sc->entry->swap_status == SWAPOUT_WRITING)
-	    assert(mem->swapout.done_offset > sc->copy_offset + mem->swap_hdr_sz);
-	file_read(sc->swapin_fd,
+	    assert(storeOffset(mem->swapout.sio) > sc->copy_offset + mem->swap_hdr_sz);
+	storeRead(sc->swapin_sio,
 	    sc->copy_buf,
 	    sc->copy_size,
 	    sc->copy_offset + mem->swap_hdr_sz,
 	    storeClientReadBody,
 	    sc);
     }
-#ifndef OPTIMISTIC_IO
-    sc->flags.disk_io_pending = 1;
-#endif
 }
 
 static void
-storeClientReadBody(int fd, const char *buf, int len, int flagnotused, void *data)
+storeClientReadBody(void *data, const char *buf, ssize_t len)
 {
     store_client *sc = data;
     MemObject *mem = sc->entry->mem_obj;
-    STCB *callback = sc->callback;
     assert(sc->flags.disk_io_pending);
     sc->flags.disk_io_pending = 0;
     assert(sc->callback != NULL);
-    debug(20, 3) ("storeClientReadBody: FD %d, len %d\n", fd, len);
+    debug(20, 3) ("storeClientReadBody: len %d\n", len);
     if (sc->copy_offset == 0 && len > 0 && mem->reply->sline.status == 0)
-	httpReplyParse(mem->reply, sc->copy_buf);
-    sc->callback = NULL;
-    callback(sc->callback_data, sc->copy_buf, len);
+	httpReplyParse(mem->reply, sc->copy_buf, headersEnd(sc->copy_buf, len));
+    storeClientCallback(sc, len);
 }
 
 static void
-storeClientReadHeader(int fd, const char *buf, int len, int flagnotused, void *data)
+storeClientReadHeader(void *data, const char *buf, ssize_t len)
 {
     store_client *sc = data;
     StoreEntry *e = sc->entry;
     MemObject *mem = e->mem_obj;
-    STCB *callback = sc->callback;
     int swap_hdr_sz = 0;
     size_t body_sz;
     size_t copy_sz;
     tlv *tlv_list;
+    tlv *t;
+    int swap_object_ok = 1;
     assert(sc->flags.disk_io_pending);
     sc->flags.disk_io_pending = 0;
     assert(sc->callback != NULL);
-    debug(20, 3) ("storeClientReadHeader: FD %d, len %d\n", fd, len);
+    debug(20, 3) ("storeClientReadHeader: len %d\n", len);
     if (len < 0) {
-	debug(20, 3) ("storeClientReadHeader: FD %d: %s\n", fd, xstrerror());
-	sc->callback = NULL;
-	callback(sc->callback_data, sc->copy_buf, len);
+	debug(20, 3) ("storeClientReadHeader: %s\n", xstrerror());
+	storeClientCallback(sc, len);
 	return;
     }
     tlv_list = storeSwapMetaUnpack(buf, &swap_hdr_sz);
+    if (swap_hdr_sz > len) {
+	/* oops, bad disk file? */
+	debug(20, 1) ("WARNING: swapfile header too small\n");
+	storeClientCallback(sc, -1);
+	return;
+    }
     if (tlv_list == NULL) {
-	debug(20, 1) ("storeClientReadHeader: failed to unpack meta data\n");
-	sc->callback = NULL;
-	callback(sc->callback_data, sc->copy_buf, -1);
+	debug(20, 1) ("WARNING: failed to unpack meta data\n");
+	storeClientCallback(sc, -1);
 	return;
     }
     /*
-     * XXX Here we should check the meta data and make sure we got
-     * the right object.
+     * Check the meta data and make sure we got the right object.
      */
+    for (t = tlv_list; t; t = t->next) {
+	switch (t->type) {
+	case STORE_META_KEY:
+	    assert(t->length == MD5_DIGEST_CHARS);
+	    if (!EBIT_TEST(e->flags, KEY_PRIVATE) &&
+		memcmp(t->value, e->key, MD5_DIGEST_CHARS)) {
+		debug(20, 1) ("WARNING: swapin MD5 mismatch\n");
+		debug(20, 1) ("\t%s\n", storeKeyText(t->value));
+		debug(20, 1) ("\t%s\n", storeKeyText(e->key));
+		swap_object_ok = 0;
+	    }
+	    break;
+	case STORE_META_URL:
+	    if (NULL == mem->url)
+		(void) 0;	/* can't check */
+	    else if (0 == strcasecmp(mem->url, t->value))
+		(void) 0;	/* a match! */
+	    else {
+		debug(20, 1) ("storeClientReadHeader: URL mismatch\n");
+		debug(20, 1) ("\t{%s} != {%s}\n", t->value, mem->url);
+		swap_object_ok = 0;
+		break;
+	    }
+	    break;
+	case STORE_META_STD:
+	    break;
+	default:
+	    debug(20, 1) ("WARNING: got unused STORE_META type %d\n", t->type);
+	    break;
+	}
+    }
     storeSwapTLVFree(tlv_list);
+    if (!swap_object_ok) {
+	storeClientCallback(sc, -1);
+	return;
+    }
     mem->swap_hdr_sz = swap_hdr_sz;
     mem->object_sz = e->swap_file_sz - swap_hdr_sz;
     /*
@@ -397,9 +456,9 @@ storeClientReadHeader(int fd, const char *buf, int len, int flagnotused, void *d
 	    copy_sz);
 	xmemmove(sc->copy_buf, sc->copy_buf + swap_hdr_sz, copy_sz);
 	if (sc->copy_offset == 0 && len > 0 && mem->reply->sline.status == 0)
-	    httpReplyParse(mem->reply, sc->copy_buf);
-	sc->callback = NULL;
-	callback(sc->callback_data, sc->copy_buf, copy_sz);
+	    httpReplyParse(mem->reply, sc->copy_buf,
+		headersEnd(sc->copy_buf, copy_sz));
+	storeClientCallback(sc, copy_sz);
 	return;
     }
     /*
@@ -410,10 +469,12 @@ storeClientReadHeader(int fd, const char *buf, int len, int flagnotused, void *d
 }
 
 int
-storeClientCopyPending(StoreEntry * e, void *data)
+storeClientCopyPending(store_client * sc, StoreEntry * e, void *data)
 {
-    /* return 1 if there is a callback registered for this client */
-    store_client *sc = storeClientListSearch(e->mem_obj, data);
+#if STORE_CLIENT_LIST_DEBUG
+    assert(sc == storeClientListSearch(e->mem_obj, data));
+#endif
+    assert(sc->entry == e);
     if (sc == NULL)
 	return 0;
     if (sc->callback == NULL)
@@ -421,54 +482,52 @@ storeClientCopyPending(StoreEntry * e, void *data)
     return 1;
 }
 
+/*
+ * This routine hasn't been optimised to take advantage of the
+ * passed sc. Yet.
+ */
 int
-storeUnregister(StoreEntry * e, void *data)
+storeUnregister(store_client * sc, StoreEntry * e, void *data)
 {
     MemObject *mem = e->mem_obj;
-    store_client *sc;
-    store_client **S;
-    STCB *callback;
+#if STORE_CLIENT_LIST_DEBUG
+    assert(sc == storeClientListSearch(e->mem_obj, data));
+#endif
     if (mem == NULL)
 	return 0;
     debug(20, 3) ("storeUnregister: called for '%s'\n", storeKeyText(e->key));
-    for (S = &mem->clients; (sc = *S) != NULL; S = &(*S)->next) {
-	if (sc->callback_data == data)
-	    break;
-    }
     if (sc == NULL)
 	return 0;
-    if (sc == mem->clients) {
+    if (mem->clients.head == NULL)
+	return 0;
+    if (sc == mem->clients.head->data) {
 	/*
 	 * If we are unregistering the _first_ client for this
 	 * entry, then we have to reset the client FD to -1.
 	 */
 	mem->fd = -1;
     }
-    *S = sc->next;
+    dlinkDelete(&sc->node, &mem->clients);
     mem->nclients--;
-    sc->flags.disk_io_pending = 0;
     if (e->store_status == STORE_OK && e->swap_status != SWAPOUT_DONE)
-	storeCheckSwapOut(e);
-    if (sc->swapin_fd > -1) {
-	commSetSelect(sc->swapin_fd, COMM_SELECT_READ, NULL, NULL, 0);
-	file_close(sc->swapin_fd);
-	store_open_disk_fd--;
-	/* XXX this probably leaks file_read handler structures */
+	storeSwapOut(e);
+    if (sc->swapin_sio) {
+	storeClose(sc->swapin_sio);
+	cbdataUnlock(sc->swapin_sio);
+	sc->swapin_sio = NULL;
+	statCounter.swap.ins++;
     }
-#if USE_ASYNC_IO
-    else
-	aioCancel(-1, sc);
-#endif
-    if ((callback = sc->callback) != NULL) {
+    if (NULL != sc->callback) {
 	/* callback with ssize = -1 to indicate unexpected termination */
 	debug(20, 3) ("storeUnregister: store_client for %s has a callback\n",
 	    mem->url);
-	sc->callback = NULL;
-	callback(sc->callback_data, sc->copy_buf, -1);
+	storeClientCallback(sc, -1);
     }
 #if DELAY_POOLS
     delayUnregisterDelayIdPtr(&sc->delay_id);
 #endif
+    cbdataUnlock(sc->callback_data);	/* we're done with it now */
+    /*assert(!sc->flags.disk_io_pending); */
     cbdataFree(sc);
     assert(e->lock_count > 0);
     if (mem->nclients == 0)
@@ -480,15 +539,21 @@ off_t
 storeLowestMemReaderOffset(const StoreEntry * entry)
 {
     const MemObject *mem = entry->mem_obj;
-    off_t lowest = mem->inmem_hi;
+    off_t lowest = mem->inmem_hi + 1;
     store_client *sc;
-    store_client *nx = NULL;
-    for (sc = mem->clients; sc; sc = nx) {
-	nx = sc->next;
+    dlink_node *nx = NULL;
+    dlink_node *node;
+
+    for (node = mem->clients.head; node; node = nx) {
+	sc = node->data;
+	nx = node->next;
 	if (sc->callback_data == NULL)	/* open slot */
 	    continue;
 	if (sc->type != STORE_MEM_CLIENT)
 	    continue;
+	if (sc->type == STORE_DISK_CLIENT)
+	    if (NULL != sc->swapin_sio)
+		continue;
 	if (sc->copy_offset < lowest)
 	    lowest = sc->copy_offset;
     }
@@ -502,17 +567,21 @@ InvokeHandlers(StoreEntry * e)
     int i = 0;
     MemObject *mem = e->mem_obj;
     store_client *sc;
-    store_client *nx = NULL;
-    assert(mem->clients != NULL || mem->nclients == 0);
+    dlink_node *nx = NULL;
+    dlink_node *node;
+
     debug(20, 3) ("InvokeHandlers: %s\n", storeKeyText(e->key));
     /* walk the entire list looking for valid callbacks */
-    for (sc = mem->clients; sc; sc = nx) {
-	nx = sc->next;
+    for (node = mem->clients.head; node; node = nx) {
+	sc = node->data;
+	nx = node->next;
 	debug(20, 3) ("InvokeHandlers: checking client #%d\n", i++);
 	if (sc->callback_data == NULL)
 	    continue;
 	if (sc->callback == NULL)
 	    continue;
+	if (sc->flags.disk_io_pending)
+	    continue;
 	storeClientCopy2(e, sc);
     }
 }
@@ -588,6 +657,6 @@ CheckQuickAbort(StoreEntry * entry)
 	return;
     if (CheckQuickAbort2(entry) == 0)
 	return;
-    Counter.aborted_requests++;
+    statCounter.aborted_requests++;
     storeAbort(entry);
 }
